#!/usr/bin/env python3
import os
import pickle
import unittest

import u4_state

TEST_DIRECTORY = os.path.dirname(__file__)


def compare_lists(x, y, test_name, db_name):
    assert len(x) == len(
        y), f'Failure while testing {test_name} with {db_name} db:\n Expected list of length {len(x)} but got {len(y)}'
    assert isinstance(x, type(
        y)), f'Failure while testing {test_name} with {db_name} db:\n Expected type {type(x)} but got {type(y)}'
    x2 = sorted(x) if all(type(g) == int for g in x) else x
    y2 = sorted(y) if all(type(g) == int for g in y) else y
    for exp, res in zip(x2, y2):
        emsg = f'Failure while testing {test_name} with {db_name} db:\n Element at index {y.index(res)}: {res} not found in expected result'
        if len(x2) < 10:
            emsg += f'exp={x2}\nres={y2}'
        assert exp == res or abs(exp-res) <= 1e-6, emsg
    return True


def compare_sets(x, y, test_name, db_name):
    assert len(x) == len(
        y), f'Failure while testing {test_name} with {db_name} db:\n Expected set of length {len(x)} but got {len(y)}'
    assert isinstance(x, type(
        y)), f'Failure while testing {test_name} with {db_name} db:\n Expected type {type(x)} but got {type(y)}'
    for item in y:
        assert item in x, f'Failure while testing {test_name} with {db_name} db:\n Item {item} not found in expected result'
    return True

def compare_dictionary(x, y, test_name, db_name):
    assert len(x) == len(
        y), f'Failure while testing {test_name} with {db_name} db:\n Expected {len(x)} number of keys but got {len(y)}'
    compare_sets(set(x.keys()), set(y.keys()), test_name, db_name)

    for key in y:
        result_values = y[key]
        expected = x[key]
        if isinstance(expected, list):
            try:
                compare_lists(expected, result_values, test_name, db_name)
            except BaseException:
                assert False, f'Failure while testing {test_name} with {db_name} db:\n Expected key {key} to have value {expected} but got {result_values}'
        if isinstance(expected, set):
            try:
                compare_sets(expected, result_values, test_name, db_name)
            except BaseException:
                assert False, f'Failure while testing {test_name} with {db_name} db:\n Expected key {key} to have value {expected} but got {result_values}'
    return True


class TestUnique(unittest.TestCase):

    def test_upper(self):
        self.assertEqual('foo'.upper(), 'FOO')

    def test_isupper(self):
        self.assertTrue('FOO'.isupper())
        self.assertFalse('Foo'.isupper())

    def test_split(self):
        s = 'hello world'
        self.assertEqual(s.split(), ['hello', 'world'])
        # check that s.split fails when the separator is not a string
        with self.assertRaises(TypeError):
            s.split(2)



    def test_unique_destinations(self):
        test_name = 'unique_destinations'
        for db_name, db in test_results['db'].items():
            expected = test_results[test_name][db_name]
            for state, exp in expected:
                result = u4_state.unique_destinations(db, state)
                self.assertEqual(result, exp)

class TestTransform(unittest.TestCase):
    def test_transform_dict(self):
        test_name = 'transform_dict'
        for db_name, db in test_results['db'].items():
            result = u4_state.transform_dict(db)
            expected = test_results[test_name][db_name]
            self.assertTrue(compare_dictionary(expected, result, test_name, db_name))

class TestRows(unittest.TestCase):
    def test_make_rows(self):
        test_name = 'make_rows'
        for db_name, db in test_results['db'].items():
            result = u4_state.make_rows(db)
            expected = test_results[test_name][db_name]
            self.assertEqual(len(result), len(expected))
            self.assertTrue(isinstance(result, type(expected)))
            for i in range(len(expected)):
                self.assertTrue(compare_lists(expected[i], result[i], test_name, db_name))


if __name__ == '__main__':
    import sys

    thismodule = sys.modules[__name__]

    filename = os.path.join(
        TEST_DIRECTORY,
        'resources',
        'tests',
        f'db_tests.pickle')
    with open(filename, 'rb') as f:
        raw = pickle.load(f)
        for db_name, db in raw['db'].items():
            setattr(thismodule, f'db_{db_name}', db)
        setattr(thismodule, 'test_results', raw)
    unittest.main()
